#!/usr/bin/python3

# *******************************************
#
# Copyright (c) 2022-2024 AO MCST
# Distributed under the terms of MIT License.
#
# *******************************************

import os, sys, string

# Check existing xorg.conf file
if os.path.exists("/etc/X11/xorg.conf"):
    if os.path.exists("/etc/X11/old.xorg.conf"):
        os.remove("/etc/X11/old.xorg.conf")
    os.rename("/etc/X11/xorg.conf", "/etc/X11/old.xorg.conf")

xorg_file = open('/etc/X11/xorg.conf', 'w')

# ____________________________________________________________
# Read an output from the terminal command (lspci, uname etc.)
# ____________________________________________________________
def read_process(terminal_command):
    process = os.popen(terminal_command, 'r', 1)
    output_string = process.read()
    process.close()

    return output_string

#_______________________________________________
# Function for parsing '\n' in the end of string
#_______________________________________________

def parse_str_end(string):
    offset = 1
    for i in range(len(string)):
        if string[-i] == '\n':
            offset += 1
        else:
            break

    return string[0:-offset]

#_________
# Get arch
#_________

def get_arch():
    arch = read_process('uname -m')

    return parse_str_end(arch)

arch = get_arch()

#______________
# Get processor
#______________

def get_processor():
    processor = read_process('uname -p')

    return parse_str_end(processor)

processor = get_processor()

# __________________________________________________________________
# Execute process from the terminal command and return it's variable
# __________________________________________________________________
def execute_process(terminal_command):
    process = os.popen(terminal_command, 'r', 1)
    return process

# _________________________________
# Get kernel driver name from lspci
# _________________________________
def get_kernel_driver(busID):
    lspci_output = read_process(f"lspci -s {busID} -k").split("\n")

    for i in range(len(lspci_output)):
        if "Kernel driver" in lspci_output[i]:
            kernel_driver_string = lspci_output[i].split(" ")

    parse_str_end(kernel_driver_string)
    return kernel_driver_string.pop()

# _________________________
# Get the name of videocard
# _________________________
def get_videocard_name(device_string):
    # Silicon Motion
    if "SM718" in device_string:
        return "SM718"
    elif "SM750" in device_string:
        return "SM750"
    elif "SM768" in device_string:
        return "SM768"
    # MCST MGA
    elif "Oracle/SUN Psycho" in device_string:
        return "MCST MGA"
    else:
        return ""

# _______________________________________________________________________________________________
# Convert device pci number from hexadecimal (lspci format) to decimal (xorg.conf "BusID" format)
# _______________________________________________________________________________________________
def xorg_convert_busid(hex_pci):
    pci = hex_pci.upper().replace(".",":").split(":")
    size = len(pci)

    if size == 3:
       bus = int(pci[0], 16)
       dev = int(pci[1], 16)
       func = int(pci[2], 16)
       pci = "{:02d}:{:02}:{:02}".format(bus, dev, func)

    elif size == 4:
       node = int(pci[0], 16)
       bus = int(pci[1], 16)
       dev = int(pci[2], 16)
       func = int(pci[3], 16)
       pci = "{:02d}@{:04}:{:02}:{:02}".format(bus, node, dev, func)

    else:
        print("BusID parsing error!")
        pci = ""

    return pci
# ____________________________________
# Get videocard list from lspci output
# ____________________________________
def find_videocards():
    lspci_output = read_process('lspci').split("\n")
    videocard_list = []

    for i in range(len(lspci_output)):
        if " VGA " in lspci_output[i] or "Display controller" in lspci_output[i]:
            videocard = get_device_info(lspci_output[i])
            videocard_list.append(videocard)

    return videocard_list

# ______________________________________________________
# Get busID, kernel driver and name of current videocard
# ______________________________________________________
def get_device_info(device_string):
    device_string_list = device_string.split(" ")

    busID_hex = device_string_list[0]         # <--- busID in hexadecimal format (from lspci)
    busID_dec = xorg_convert_busid(busID_hex) # <--- busID in decimal format (used in xorg.conf)

    kernel_driver = get_kernel_driver(busID_hex)

    videocard_name = get_videocard_name(device_string_list)

    videocard = {'busID': busID_dec, 'kernel_driver': kernel_driver, 'name': videocard_name}

    return videocard

# ____________________________________________________________
#  Get monitors list from the xrandr output (currently unused)
# ____________________________________________________________
def find_monitors():
    xrandr_output = get_xrandr_output()
    monitor_list = []

    for i in range(len(xrandr_output)):
        if "HDMI" in xrandr_output[i] or "VGA" in xrandr_output[i] \
        or "DVI" in xrandr_output[i] or "DisplayPort" in xrandr_output[i]:
            if " connected " in xrandr_output[i]:
                monitor = get_monitor_info(xrandr_output[i])
                monitor_list.append(monitor)

    return monitor_list

# ________________________________________________
#  Check if X is running and get the xrandr output
# ________________________________________________
def get_xrandr_output():
    X_is_running = read_process('pidof X')
    Xorg_is_running = read_process('pidof Xorg')
    if not X_is_running and not Xorg_is_running:
        X_process = execute_process('X :0 > /dev/null 2>&1')
        #X_process = Process(target=execute_process, args=('X :0 > /dev/null 2>&1',))
        #X_process.start()
    else:
        X_process = False

    xrandr_output = read_process('DISPLAY=:0 xrandr -q').split("\n")
    if X_process:
        X_process.close()

    return xrandr_output

# _________________________________________
#  Get monitor's name from xrandr string
# _________________________________________
def get_monitor_info(monitor_string):
    monitor_string_list = monitor_string.split(" ")

   # Monitor's name is always the first word in xrandr string (HDMI-0, VGA-0 etc.)
    return monitor_string_list[0]

# ______________________________________
# Print graphics device section's header
# ______________________________________
def print_device_section_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Graphics device section\n' +
                    '# ****************************************************************\n\n' +
                    '# Devices configured by xorgconfig:\n')

# _________________________________________________________________________________________
# Fill "Device" sections in xorg.conf (additional_options is a string or a list of strings)
# _________________________________________________________________________________________
def fill_device_section(device_number, driver, busID, additional_options):
    xorg_file.write( '\nSection "Device"\n' +
                    f'    Identifier   "Device {device_number}"\n' +
                    f'    Driver       "{driver}"\n' +
                    f'    BusID        "PCI:{busID}"\n')

    if additional_options:
        write_additional_options(additional_options)

    xorg_file.write('EndSection\n')

# ____________________________________________
# Write additional options in "Device" section
# ____________________________________________
def write_additional_options(options):
    if type(options) is str:
        xorg_file.write(f'    Option       {options}\n')

    elif type(options) is list:
        for i in range(len(options)):
            xorg_file.write(f'    Option       {options[i]}\n')

# ____________
# Videocards
# ____________
def do_radeon(device_number, videocard):
    busID = videocard['busID']
    fill_device_section(device_number, 'radeon', busID, '')


def do_amdgpu(device_number, videocard):
    busID = videocard['busID']
    fill_device_section(device_number, 'amdgpu', busID, '')


def do_silicon_motion(device_number, videocard):
    busID = videocard['busID']
    name = videocard['name']
    additional_option = ''

    if name == "SM768":
        fill_device_section(device_number, 'modesetting', busID, additional_option)

    elif name == "SM718" or name == "SM750":
        if arch == "sparc" or arch == "sparc64":
            fill_device_section(device_number, 'fbdev', busID, additional_option)
        else:
            additional_option = '"MigrationHeuristic" "greedy"'
            fill_device_section(device_number, 'smi', busID, additional_option)


def do_aspeed(device_number, videocard):
    busID = videocard['busID']
    fill_device_section(device_number, 'modesetting', busID, '')


def do_nvidia(device_number, videocard):
    busID = videocard['busID']
    fill_device_section(device_number, 'nouveau', busID, '')


def do_mcst_mga(device_number, videocard):
    busID = videocard['busID']
    fill_device_section(device_number, 'mcstmga', busID, '')
    # TODO

def do_mga(device_number, videocard):
    busID = videocard['busID']
    if processor == "E1C+" or arch == "R2000+":
        fill_device_section(device_number, 'mga2', busID, '')

    elif processor == "E2C3": # "pvr" driver in future
        fill_device_section(device_number, 'modesetting', busID, '')

    else:
        print("Error: you can't use MGA2 driver on this machine")

# ____________________________________________________________
# Check videocards and fill "Device" sections in xorg.conf
# ____________________________________________________________
def do_device(videocard_list):
    print_device_section_header()

    for i in range(len(videocard_list)):
        kernel_driver = videocard_list[i]['kernel_driver']

        if kernel_driver == "radeon":
            do_radeon(i, videocard_list[i])
        elif kernel_driver == "amdgpu":
            do_amdgpu(i, videocard_list[i])
        elif kernel_driver == "lynxfb" or kernel_driver == "smifb": # Silicon Motion
            do_silicon_motion(i, videocard_list[i])
        elif kernel_driver == "ast": # ASPEED
            do_aspeed(i, videocard_list[i])
        elif kernel_driver == "nouveau": # Nvidia
            do_nvidia(i, videocard_list[i])
        elif kernel_driver == "mgam83fb": # MCST MGA
            do_mcst_mga(i, videocard_list[i])
        elif kernel_driver == "mga2" or kernel_driver == "mga25":
            do_mga(i, videocard_list[i])
        else:
            print("Warning: xorg.conf was created without BusID and Driver because didn't find known equipment")


# ____________________________________
#  Fill "Module" section in xorg.conf
# ____________________________________
def print_module_section_header():
    xorg_file.write('# ****************************************************************\n' +
                    '# Modules section\n' +
                    '# ****************************************************************\n')

def fill_module_section():
    xorg_file.write( '\nSection "Module"\n' +
                     '    Load    "dbe"   # Double buffer extension\n\n' +
                     '    SubSection  "extmod"\n' +
                     '        Option  "omit xfree86-dga"   # Don\'t initialise the DGA extension\n' +
                     '    EndSubSection\n\n' +
                     '    Load    "type1"\n' +
                     '    Load    "freetype"\n' +
                     '    Load    "glx"\n' +
                     '    Load    "dri"\n' +
                     'EndSection\n')

def do_module():
    print_module_section_header()
    fill_module_section()

# _________________________________________________________
#  Check fonts' paths and fill "Files" section in xorg.conf
# _________________________________________________________
def print_files_section_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Files section.  This allows default font and rgb paths to be set\n' +
                    '# ****************************************************************\n')

def fill_files_section():
    font_prefix = "/usr/share/fonts/X11"
    font = ["cyrillic", "misc", "TTF", "Type1", "75dpi", "100dpi", "75dpi/:unscaled", "100dpi/:unscaled"]

    xorg_file.write( '\nSection "Files"\n')
    for i in range(len(font)):
        xorg_file.write(f'    FontPath    "{font_prefix}/{font[i]}"\n')
    xorg_file.write( 'EndSection\n')

def do_files():
    print_files_section_header()
    fill_files_section()

# _________________________________________
#  Fill "ServerFlags" section in xorg.conf
# _________________________________________
def print_server_flags_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Server flags section.\n' +
                    '# ****************************************************************\n')

def fill_server_flags_section():
    xorg_file.write( '\nSection "ServerFlags"\n' +
                     '    Option "AutoAddDevices" "True"\n' +
                     'EndSection\n')

def do_server_flags():
    print_server_flags_header()
    fill_server_flags_section()

# _________________________________________
#  Fill "InputDevice" section in xorg.conf
# _________________________________________
def print_input_device_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Input devices.\n' +
                    '# ****************************************************************\n')

def fill_mouse():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Core Pointer\'s InputDevice section\n' +
                    '# ****************************************************************\n')

    xorg_file.write( '\nSection "InputDevice"\n' +
                     '    Identifier  "Mouse1"\n' +
                     '    Driver      "mouse"\n' +
                     '    Option      "CorePointer"\n' +
                     '    Option      "Protocol"     "ImPS/2"\n' +
                     '    Option      "Device"       "/dev/psaux"\n' +
                     '    Option      "ZAxisMapping" "4 5 6 7"\n' +
                     'EndSection\n')

def fill_elo():
    # TODO
    return

def fill_input_device_section():
    fill_mouse()
    fill_elo()

def do_input_device():
    print_input_device_header()
    fill_input_device_section()


# _____________________________________
#  Fill "Monitor" section in xorg.conf
# _____________________________________
def print_monitor_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Monitor section\n' +
                    '# ****************************************************************\n')

def fill_monitor_section(videocard_list):
    # Create special Monitor section for every videcard
    for videocard_counter in range(len(videocard_list)):
        xorg_file.write( '\nSection "Monitor"\n' +
                        f'    Identifier  "Monitor {videocard_counter}"\n')

        # Insert special options for MCST MGA videocard
        if videocard_list[videocard_counter]['name'] == "MCST MGA":
            xorg_file.write('    HorizSync   30 - 70\n' +
                            '    VertRefresh 50-160\n' +
                            '    Modeline "1024x768" 64.994 1024 1064 1200 1344  768 771 777 806  -HSync -VSync\n' +
                            '    Modeline "1024x768@85" 98.9 1024 1056 1216 1408 768 782 788 822 -HSync -VSync\n' +
                            '    Modeline "1024x768@100" 115.5 1024 1056 1248 1440 768 771 781 802 -HSync -VSync\n' +
                            '    Modeline "1920x1080" 148.550  1920 2048 2080 2200  1080 1096 1100 1125  +Hsync +Vsync\n')

        xorg_file.write( 'EndSection\n')

def do_monitor(videocard_list):
    print_monitor_header()
    fill_monitor_section(videocard_list)

# ______________________________________________________
#  Check screens and fill "Screen" sections in xorg.conf
# ______________________________________________________
def print_screen_header():
    xorg_file.write('\n# ****************************************************************\n' +
                    '# Screen sections\n' +
                    '# ****************************************************************\n')

def fill_display_subsection():
    depth = [1, 4, 8, 15, 16, 24]
    for i in range(len(depth)):
        xorg_file.write( '    SubSection "Display"\n' +
                         '            Viewport   0 0\n' +
                        f'            Depth     {depth[i]}\n' +
                         '    EndSubSection\n')

def fill_screen_section(videocard_list):
    for i in range(len(videocard_list)):
         xorg_file.write( '\nSection "Screen"\n' +
                         f'    Identifier  "Screen {i}"\n' +
                         f'    Device      "Device {i}"\n' +
                         f'    Monitor     "Monitor {i}"\n')
         fill_display_subsection()
         xorg_file.write('EndSection\n')

def do_screen(videocard_list):
    print_screen_header()
    fill_screen_section(videocard_list)

# _______________________________________
#  Fill "Extensions" section in xorg.conf
# _______________________________________
def do_extensions():
    xorg_file.write( '\nSection "Extensions"\n' +
                     '    Option "Composite" "Enable"\n' +
                     'EndSection\n')

# _____________________________________
#  Fill "dri" section in xorg.conf
# _____________________________________
def do_dri():
    xorg_file.write( '\nSection "dri"\n' +
                     '    Mode 0666\n'
                     'EndSection\n')

# ______________________________
# Fill the sections in xorg.conf
# ______________________________
def fill_xorg_conf(videocard_list):
    do_module()
    do_files()
    do_server_flags()
    do_input_device()
    do_monitor(videocard_list)
    do_device(videocard_list)
    do_screen(videocard_list)
    do_extensions()
    do_dri()


# MAIN
videocard_list = find_videocards()

fill_xorg_conf(videocard_list)

xorg_file.close()
